/*
 * DNG processings
 *  Copyright (C) 2015-2017 Noriaki TANAKA (dtana@startide.jp)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Id: dng.c,v 1.60 2017/02/12 10:22:59 dtana Exp $
 */
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <math.h>

#include "tiffmd.h"
#include "dng.h"

int verbose;

#define MAX_DIMN 64

struct dim {
	int c, r, v[MAX_DIMN];
};

struct area {
	int x, y, w, h;
};

struct dng {
	int w, h, bps;
	long offs, size;
	struct tiffmd *ip;

	void *meta; /* buffer of <offs> bytes */
	uint16_t *data; /* rest as raster */

	struct dim cfac, bl;
	struct area active;
};

int
dng_w(const struct dng *p)
{
	return p->w;
}

int
dng_h(const struct dng *p)
{
	return p->h;
}

static int tiffmd_intv(const struct tiffmd *, const char *);
static int expand_dimv(struct dng *,
    const char *, const char *, struct dim *);
static int expand_area(struct dng *, const char *, struct area *);


static const struct tiffmd *find_base_image(const struct tiffmd *);
static void show_hist(const struct dng *);

static int freadpx(uint16_t *, int, int, FILE *),
	fwritepx(const uint16_t *, int, int, FILE *);

static void
efclose(FILE *fp)
{
	int errno_ = errno;
	fclose(fp), errno = errno_;
}

struct dng *
load_dng(const char *path)
{
	struct tiffmd *ip;
	const struct tiffmd *bp;
	int w, h;
	int strip_offs, strip_size;
	struct dng *dp;
	FILE *fp;
	
	if ((fp = fopen(path, "rb")) == NULL)
		return NULL;

	bp = NULL;
	if ((ip = load_tiffmd(fp)) == NULL) {
		fprintf(stderr, "TIFF header not found\n");
		goto error_;
	}
	if ((bp = find_base_image(ip)) == NULL) {
		fprintf(stderr, "DNG base image not found\n");
		goto error_;
	}
	
	w = tiffmd_intv(bp, "IMAGEWIDTH");
	h = tiffmd_intv(bp, "IMAGELENGTH");
	if (tiffmd_intv(bp, "COMPRESSION") != 1) {
		fprintf(stderr, "non compression only\n");
		errno = EINVAL;
		goto error_;
	}
	if (tiffmd_intv(bp, "BITSPERSAMPLE") != 16) {
		fprintf(stderr, "assumed 16 bits only\n");
		errno = EINVAL;
		goto error_;
	}
	if ((strip_offs = tiffmd_intv(bp, "STRIPOFFSETS")) <= 0) {
		fprintf(stderr, "invalid strip offset\n");
		errno = EINVAL;
		goto error_;
	}
	if ((strip_size = tiffmd_intv(bp, "STRIPBYTECOUNTS")) <= 0) {
		fprintf(stderr, "invalid strip byte counts\n");
		errno = EINVAL;
		goto error_;
	}
	
	if (w * h * 2 != strip_size) {
		fprintf(stderr, "strip size missmatch: %d, %d\n",
		    w * h * 2, strip_size);
		errno = EINVAL;
		goto error_;
	}
	
	if ((dp = malloc(sizeof(struct dng))) == NULL)
		goto error_;
	dp->w = w;
	dp->h = h;
	dp->offs = strip_offs;
	dp->size = strip_size;
	dp->ip = ip;

	dp->bps = dng_estimated_bps(dp);

	if (expand_area(dp, "ActiveArea", &dp->active) < 0) {
		fprintf(stderr, "No ActiveArea tag\n");
		dp->active.x = dp->active.y = 0;
		dp->active.w = dp->w;
		dp->active.h = dp->h;
	}

	if (expand_dimv(dp,
		"CFARepeatPatternDim", "CFAPattern", &dp->cfac) < 0) {
		fprintf(stderr, "cannot read CFA pattern\n");
		/* assumed RGGB 2-dimensional */
		dp->cfac.c = dp->cfac.r = 2;
		dp->cfac.v[0] = 0;
		dp->cfac.v[1] = 1;
		dp->cfac.v[2] = 1;
		dp->cfac.v[3] = 2;
	}

	if (expand_dimv(dp,
		"BlackLevelRepeatDim", "BlackLevel", &dp->bl) < 0) {
		dp->bl.c = dp->bl.r = 1;
		dp->bl.v[0] = 0;
	}
	
	dp->meta = dp->data = NULL;

	rewind(fp);
	if ((dp->meta = malloc(dp->offs)) == NULL) {
		perror("malloc");
		goto error_;
	}
	if (fread(dp->meta, (size_t)dp->offs, 1, fp) != 1) {
		perror(path);
		goto error_;
	}
	
	if ((dp->data = malloc(dp->size)) == NULL) {
		perror("malloc");
		goto error_;
	}
	if (freadpx(dp->data, dp->w * dp->h, tiffmd_xo(dp->ip), fp) < 0)
		goto error_;

	fclose(fp);

	if (verbose) {
		printf("%d x %d\n", w, h);
		printf("%d : %d bytes\n", strip_offs, strip_size);
		printf("%d bits\n", dp->bps);
		show_hist(dp);
	}
	return dp;

error_:
	term_tiffmd(ip);
	efclose(fp);
	return NULL;
}

void
term_dng(struct dng *p)
{
	if (p != NULL) {
		free(p->meta);
		free(p->data);
		term_tiffmd(p->ip);
	}
	free(p);
}

int
dng_bps(const struct dng *p)
{
	return p->bps;
}

static const struct tiffmd *
find_base_image(const struct tiffmd *p)
{
	int k;
	
	if (tiffmd_intv(p, "NEWSUBFILETYPE") == 0)
		return p;
	for (k = 0;; k++) {
		const struct tiffmd *sp;

		if ((sp = find_tiffmd_node(p,
			    tiffmd_tag("SUBIFDS"), k)) == NULL)
			break;
		if (tiffmd_intv(sp, "NEWSUBFILETYPE") == 0)
			return sp;
	}
	return NULL;
}

static int
getn(int *r, const struct tiffmd *mp, int k)
{
	int c[2];
	
	switch (tiffmd_getn(c, 2, mp, k)) {
	case 1:
		*r = c[0];
		break;
	case 2:
		*r = c[0] / c[1];
		break;
	default:
		return -1;
	}
	return 0;
}

static int
get_dimv(const struct dng *p,
    const struct dim *dp, int x, int y)
{
	if (x < 0 || p->w <= x || y < 0 || p->h <= y)
		return -1;
	
	x = (x - p->active.x) % dp->c;
	y = (y - p->active.y) % dp->r;
	return dp->v[y * dp->c + x];
}

static int
expand_dimv(struct dng *p,
    const char *td, const char *tv, struct dim *dp)
{
	int x, y;
	const struct tiffmd *mp;
	
	if ((mp = find_tiffmd(p->ip, tiffmd_tag(td))) == NULL)
		return -1;

	if (tiffmd_getn(&dp->c, 1, mp, 0) != 1 ||
	    tiffmd_getn(&dp->r, 1, mp, 1) != 1)
		return -1;

	if (MAX_DIMN <= dp->c * dp->r) {
		fprintf(stderr, "%s: number exceeded\n", td);
		return -1;
	}
	
	if ((mp = find_tiffmd(p->ip, tiffmd_tag(tv))) == NULL)
		return -1;
	for (y = 0; y < dp->r; y++)
		for (x = 0; x < dp->c; x++) {
			int n, r;
			
			n = y * dp->c + x;
			if (getn(&r, mp, n) < 0)
				return -1;
			dp->v[n] = r;
		}
	
	return 0;
}

static int
expand_area(struct dng *p, const char *tn, struct area *ap)
{
	const struct tiffmd *mp;

	if ((mp = find_tiffmd(p->ip, tiffmd_tag(tn))) == NULL)
		return -1;
	if (tiffmd_getn(&ap->x, 1, mp, 1) != 1 ||
	    tiffmd_getn(&ap->y, 1, mp, 0) != 1 ||
	    tiffmd_getn(&ap->w, 1, mp, 3) != 1 ||
	    tiffmd_getn(&ap->h, 1, mp, 2) != 1) {
		return -1;
	}
	/* fprintf(stderr, "%d %d %d %d\n", ap->x, ap->y, ap->w, ap->h); */
	return 0;
}

enum DNG_CF
dng_cfac(const struct dng *p, int x, int y)
{
	return get_dimv(p, &p->cfac, x, y);
}

int
dng_black_level(const struct dng *p, int x, int y)
{
	return get_dimv(p, &p->bl, x, y);
}

#define BL(p,x,y) dng_black_level((p),(x),(y))

int
dng_get(const struct dng *p, int x, int y)
{
	if (x < 0 || p->w <= x || y < 0 || p->h <= y)
		return -1;
	return p->data[y * p->w + x];
}

void
dng_put(struct dng *p, int x, int y, unsigned u)
{
	p->data[y * p->w + x] = u;
}

/* TODO: currently, print only max/min values */
static void
show_hist(const struct dng *p)
{
	int x, y;
	int vmin, vmax;
	
	vmax = 0, vmin = 0xffff;
	for (y = 0; y < p->h; y++)
		for (x = 0; x < p->w; x++) {
			unsigned pv = (unsigned)dng_get(p, x, y);
			if (vmax < pv)
				vmax = pv;
			if (pv < vmin)
				vmin = pv;
		}
	fprintf(stderr, "%x - %x\n", vmin, vmax);
}

int
dng_estimated_bps(const struct dng *p)
{
	int white_level;
	
	if ((white_level = tiffmd_intv(p->ip, "WHITELEVEL")) < 0)
		return 16;

	if (0x4000 <= white_level)
		return 16;
	if (0x1000 <= white_level)
		return 14;
	if (0x0100 <= white_level)
		return 12;
	return 8;
}

static unsigned
bps_to_maxv(int bps)
{
	if (bps ==  8)
		return 0x00ff;
	if (bps == 12)
		return 0x0fff;
	if (bps == 14)
		return 0x3fff;
	return 0xffff;
}

static void
disabled_tag(struct dng *dp, const char *name)
{
	struct tiffmd *p;
	int tag;
	
	if ((tag = tiffmd_tag(name)) < 0)
		fprintf(stderr, "internal, %s: Unknown tag, abort\n", name), 
		    abort();
	if ((p = (struct tiffmd *)find_tiffmd(dp->ip, tag)) == NULL) {
		if (verbose)
			fprintf(stderr, "%s: not found, ignored\n", name);
		return;
	}
	/* set tiffmd to set as harmless "ImageDescription" tag */
	tiffmd_set_disable(p);
	if (tiffmd_pute(p, dp->meta, dp->offs) < 0)
		fprintf(stderr, "pointer overflow\n");
}

static void
chop_maxv(struct dng *dp)
{
	int x, y;
	unsigned maxv;
	
	maxv = bps_to_maxv(dp->bps);
	for (y = 0; y < dp->h; y++)
		for (x = 0; x < dp->w; x++)
			if (maxv < (unsigned)dng_get(dp, x, y))
				dng_put(dp, x, y, maxv);
}

int
save_dng(const struct dng *dp, const char *path)
{
	FILE *fp;
	
	if ((fp = fopen(path, "wb")) == NULL)
		return -1;

	/* to supress "broken image" warnings */
	disabled_tag((struct dng *)dp, "RawDataUniqueID");
	disabled_tag((struct dng *)dp, "NewRawImageDigest");
	if (fwrite(dp->meta, dp->offs, 1, fp) != 1)
		goto error_;
	
	chop_maxv((struct dng *)dp);

	if (fwritepx(dp->data,
		dp->w * dp->h, tiffmd_xo(dp->ip), fp) < 0)
		goto error_;
	
	return fclose(fp);
error_:
	efclose(fp);
	return -1;
}

static int
fwritepx(const uint16_t *data, int n, int xo, FILE *fp)
{
	if (xo)
		for (; n--; data++) {
			char b[2];

			b[0] = ((const char *)data)[1];
			b[1] = ((const char *)data)[0];
			if (fwrite(b, 2, 1, fp) != 1)
				return -1;
		}
	else
		if (fwrite(data, sizeof(*data), n, fp) != n)
			return -1;
	return 0;
}

static int
freadpx(uint16_t *data, int n, int xo, FILE *fp)
{
	if (xo)
		for (; n--; data++) {
			char b[2];
			
			if (fread(b, 2, 1, fp) != 1)
				return -1;
			((char *)data)[0] = b[1];
			((char *)data)[1] = b[0];
		}
	else
		if (fread(data, sizeof(*data), n, fp) != n)
			return -1;
	return 0;
}

int
dng_metadata_get_intv(const struct dng *p, const char *tc)
{
	return tiffmd_intv(p->ip, tc);
}

static int
tiffmd_intv(const struct tiffmd *p, const char *tn)
{
	int r, tag;
	const struct tiffmd *sp;

	if ((tag = tiffmd_tag(tn)) < 0)
		fprintf(stderr, "%s: unknown tag\n", tn), exit(1);

	if ((sp = find_tiffmd(p, tag)) == NULL)
		return -1;
	if (tiffmd_getn(&r, 1, sp, 0) != 1)
		return -1;
	return r;
}

int
dng_metadata_set_intv(struct dng *p, const char *tc, int v)
{
	int tag;
	struct tiffmd *mp;
	
	if ((tag = tiffmd_tag(tc)) < 0) {
		fprintf(stderr, "%s: unknown tag, ignored\n", tc);
		return -1;
	}
	if ((mp = (struct tiffmd *)find_tiffmd(p->ip, tag)) == NULL) {
		fprintf(stderr, "%s: not found\n", tc);
		return -1;
	}
	tiffmd_setn(&v, 1, mp, 0);
	tiffmd_pute(mp, p->meta, p->offs);

	return 0;
}

/* calc arithmetic mean on bayer pixels */
struct dng_average {
	int np, nf;
	uint32_t *data;
};

struct dng_average *
dng_average_init(const struct dng *bp)
{
	struct dng_average *ap;
	int k;
	
	if ((ap = malloc(sizeof(struct dng_average))) == NULL)
		return NULL;
	ap->np = bp->w * bp->h;
	if ((ap->data = malloc(sizeof(*ap->data) * ap->np)) == NULL) {
		free(ap);
		return NULL;
	}
	for (k = 0; k < ap->np; k++)
		ap->data[k] = bp->data[k];
	ap->nf = 1;

	return ap;
}

int
dng_average_add(struct dng_average *ap, const struct dng *tp)
{
	int k;
	
	if (tp->w * tp->h != ap->np) {
		fprintf(stderr, "n of pixels mismatch\n");
		return -1;
	}
	for (k = 0; k < ap->np; k++)
		ap->data[k] += tp->data[k];
	ap->nf++;
	return 0;
}

int
dng_average_calc(const struct dng_average *ap, struct dng *bp)
{
	int k;
	
	if (bp->w * bp->h != ap->np) {
		fprintf(stderr, "number of pixels mismatch\n");
		return -1;
	}
	for (k = 0; k < ap->np; k++)
		bp->data[k] = ap->data[k] / ap->nf;
	return 0;
}

void
dng_average_term(struct dng_average *p)
{
	if (p != NULL)
		free(p->data);
	free(p);
}
