/*
 * spot noise reduction
 *  Copyright (C) 2015-2017 Noriaki TANAKA (dtana@startide.jp)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Id: spot.c,v 1.9 2017/02/12 10:22:59 dtana Exp $
 */
#include <stdio.h>
#include <math.h>
#include "dng.h"

static void spot_fix_(struct dng *, const struct dng *, enum DNG_CF, int);

int
dng_spot_fix_with_dark(struct dng *p, const struct dng *dark)
{
	spot_fix_(p, dark, R, 40);
	spot_fix_(p, dark, G, 20);
	spot_fix_(p, dark, B, 40);
	
	return 0;
}

static double calc_ch_average(const struct dng *, enum DNG_CF);
static double calc_ch_deviation(const struct dng *, enum DNG_CF, double);

static unsigned comp_with_near_px(const struct dng *, int, int);

static void
spot_fix_(struct dng *p, const struct dng *dark, enum DNG_CF ch, int ns)
{
	double mu, sigma;
	int x, y, w, h;
	
	mu = calc_ch_average(dark, ch);
	sigma = calc_ch_deviation(dark, ch, mu);

	w = dng_w(p), h = dng_h(p);
	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++) {
			if (dng_cfac(p, x, y) != ch)
				continue;
			if (dng_get(dark, x, y) < mu + ns * sigma)
				continue;
			dng_put(p, x, y, comp_with_near_px(p, x, y));
		}
}

static unsigned
comp_with_near_px(const struct dng *p, int cx, int cy)
{
	int x, y, np, rp, dr;
	enum DNG_CF ch;

	ch = dng_cfac(p, cx, cy);
	dr = ch == G ? 1 : 2;
	rp = 0, np = 0;
	for (y = cy - dr; y <= cy + dr; y++)
		for (x = cx - dr; x <= cx + dr; x++) {
			int tp;

			if (x == cx && y == cy)
				continue;
			if (dng_cfac(p, x, y) != ch)
				continue;
			if ((tp = dng_get(p, x, y)) < 0)
				continue;
			rp += tp, np++;
		}
	return rp / np;
}

static double
calc_ch_average(const struct dng *p, enum DNG_CF ch)
{
	int x, y, n, w, h;
	double sum;

	sum = 0.0, n = 0;
	w = dng_w(p), h = dng_h(p);
	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++) {
			if (dng_cfac(p, x, y) != ch)
				continue;
			sum += dng_get(p, x, y), n++;
		}
	return sum / n;
}

static double
calc_ch_deviation(const struct dng *p, enum DNG_CF ch, double mu)
{
	int x, y, n, w, h;
	double sum;

	sum = 0.0, n = 0;
	w = dng_w(p), h = dng_h(p);
	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++) {
			double ms;
			
			if (dng_cfac(p, x, y) != ch)
				continue;
			ms = mu - dng_get(p, x, y);
			sum += ms * ms, n++;
		}
	return sqrt(sum / n);
}

static void
print_stat(const struct dng *p, int c, int ns)
{
	int x, y, w, h;
	double mu, sigma;
	
	mu = calc_ch_average(p, c);
	sigma = calc_ch_deviation(p, c, mu);
	
	printf("%s: mu: %f, sigma: %f\n",
	    (c == R)? "R": ((c == G) ? "G": "B"), mu, sigma);

	w = dng_w(p), h = dng_h(p);
	for (y = 0; y < h; y++)
		for (x = 0; x < w; x++) {
			double pv;

			if (dng_cfac(p, x, y) != c)
				continue;
			pv = dng_get(p, x, y);
			if (pv < mu - ns * sigma) {
				printf("%4d, %4d: cold: %d\n",
				    x, y, (int)pv);
			}
			if (mu + ns * sigma < pv) {
				printf("%4d, %4d: hot:  %d\n",
				    x, y, (int)pv);
			}
		}
}

int
dng_dump_spots(const struct dng *p)
{
	print_stat(p, R, 40);
	print_stat(p, G, 20);
	print_stat(p, B, 40);

	return 0;
}

/* alternative spot fixes */
int verbose;

static int
get_neighbor_pixs(unsigned u[],
    const struct dng *p, int xc, int yc, int ndim)
{
	int k, x, y, ch;
	
	ch = dng_cfac(p, xc, yc);
	
	k = 0;
	for (y = yc - ndim; y <= yc + ndim; y++)
		for (x = xc - ndim; x <= xc + ndim; x++) {
			int tp;
			if (x == xc && y == yc)
				continue;
			if (dng_cfac(p, x, y) != ch)
				continue;
			if ((tp = dng_get(p, x, y)) < 0)
				continue;
			u[k++] = (unsigned)tp;
		}
	return k;
}

#define AVERAGE_DIM 2

static double
average(const unsigned u[], int n)
{
	int k;
	double sum;
	    
	sum = 0.0;
	for (k = 0; k < n; k++)
		sum += u[k];
	return sum / n;
}

static double
deviation(const unsigned u[], int n, double mu)
{
	int k;
	double sum;
	
	sum = 0.0;
	for (k = 0; k < n; k++) {
		double ms = mu - u[k];
		sum += ms * ms;
	}
	return sqrt(sum / n);
}

#define UBUFF_SIZE(dim) (((dim)*2+1)*((dim)*2+1))
#define NSIGMA 4

static int
spot_fix(struct dng *p, int x, int y)
{
	unsigned u[UBUFF_SIZE(AVERAGE_DIM)], uc, ur, ul;
	double mu, sg;
	int n;
	
	ul = dng_get(p, x - 2, y);
	uc = dng_get(p, x,     y);
	ur = dng_get(p, x + 2, y);
	if (ur < ul) {
		unsigned ut;
		ut = ur, ur = ul, ul = ut;
	}
	if (ul < uc && uc < ur)
		return 0;

	n = get_neighbor_pixs(u, p, x, y, AVERAGE_DIM);

	mu = average(u, n);
	sg = deviation(u, n, mu);

	if (uc < mu - NSIGMA * sg) {
		if (verbose)
			fprintf(stderr, "D (%4d, %4d) %3.0f %3.0f %3d\n",
			    x, y, mu, sg, uc);
		dng_put(p, x, y, (unsigned)mu);
		return 1;
	}
	if (mu + NSIGMA * sg < uc) {
		if (verbose)
			fprintf(stderr, "L (%4d, %4d) %3.0f %3.0f %3d\n",
			    x, y, mu, sg, uc);
		dng_put(p, x, y, (unsigned)mu);
		return 1;
	}
	return 0;
}

int
dng_spot_fix(struct dng *p)
{
	int x, y, w, h, n;
	
	w = dng_w(p), h = dng_h(p);
	for (y = 0, n = 0; y < h; y++)
		for (x = 2; x < w - 2; x++)
			n += spot_fix(p, x, y);
	fprintf(stderr, "%d spots\n", n);
	return n;
}
