/*
 * srawp - Sensor Raw Preprocessor for Astrophotography
 *  Copyright (C) 2015-2017 Noriaki TANAKA (dtana@startide.jp)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Id: srawp.c,v 1.28 2017/02/12 10:22:59 dtana Exp $
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "dng.h"

void
usage(void)
{
	fprintf(stderr,
	    "Usage: srawp [-d dark.dng][-f flat.dng][-s]"
	    "[-o <output>] <light...>\n");
	fprintf(stderr,
	    "     dark.dng ... dark frame "
	    "with same ISO and exposure time without NR\n"
	    "     flat.dng ... flat frame\n");
	exit(1);
}

int verbose = 0;
int statist = 0;
int spotfix = 0;
int channel = -1;

int bias[3] = {0, 0, 0};

const char *path_dark = NULL;
const char *path_flat = NULL;
const char *path_clut = NULL;

extern int getopt(), optind;
extern char *optarg;

static struct dng *proc(const char *);

static void
extract_bias(int b[3], const char *s)
{
	if (sscanf(s, "%d,%d,%d", &b[0], &b[1], &b[2]) == 3)
		return;
	fprintf(stderr,
	    "set bias value as <r,g,b> with comma separation\n");
	exit(1);
}

static int
parse_ch(const char *s)
{
	int c = toupper(s[0]);
	if (c == 'R')
		return R;
	if (c == 'G')
		return G;
	if (c == 'B')
		return B;
	return -1;
}

int
main(int argc, char **argv)
{
	int optc;
	struct dng *bp;
	const char *path_output = NULL;
	
	verbose = 0;
	while ((optc = getopt(argc, argv, "o:d:f:l:b:c:sVS")) != EOF)
		switch (optc) {
		case 'o':
			path_output = optarg;
			break;
		case 'd':
			path_dark = optarg;
			break;
		case 'f':
			path_flat = optarg;
			break;
		case 'l':
			path_clut = optarg;
			break;
		case 'b':
			extract_bias(bias, optarg);
			break;
		case 's':
			spotfix = 1;
			break;
		case 'c':
			if ((channel = parse_ch(optarg)) < 0)
				usage();
			break;
		case 'V':
			verbose = 1;
			break;
		case 'S':
			statist = 1;
			break;
		default:
			usage();
		}
	if (optind == argc)
		usage();

	bp = proc(argv[optind++]);
	if (optind != argc) {
		struct dng_average *ap;
		
		if (verbose)
			fprintf(stderr,
			    "calc arithmetic average of %d frames\n",
			    argc - optind + 1);
		
		if ((ap = dng_average_init(bp)) == NULL)
			perror("init_average"), exit(1);
		
		while (optind < argc) {
			struct dng *tp;
			const char *path;
			
			path = argv[optind++];
			if ((tp = proc(path)) == NULL)
				perror(path), exit(1);
			dng_average_add(ap, tp);
			term_dng(tp);
		}
		
		dng_average_calc(ap, bp);
		dng_average_term(ap);
	}
	
	if (path_output != NULL) {
		if (verbose)
			fprintf(stderr, "%s writing\n", path_output);
		if (save_dng(bp, path_output) < 0)
			perror(path_output), exit(1);
	}
	
	return 0;
}

static struct dng *
proc(const char *path)
{
	struct dng *base, *dark;
	
	if ((base = load_dng(path)) == NULL)
		perror(path), exit(1);
	if (statist)
		dng_dump_spots(base);
	
	if (path_dark != NULL) {
		if ((dark = load_dng(path_dark)) == NULL)
			perror(path_dark), exit(1);
		if (dng_dark(base, dark) < 0)
			fprintf(stderr, "abort dark processing\n"), exit(1);
	} else
		dark = NULL;
	if (path_flat != NULL) {
		struct dng *flat;
		
		if ((flat = load_dng(path_flat)) == NULL)
			perror(path_flat), exit(1);
		if (dng_flat(base, flat, bias) < 0)
			fprintf(stderr, "abort flat processing\n"), exit(1);
		term_dng(flat);
	}
	if (spotfix) {
		if (dark != NULL)
			dng_spot_fix_with_dark(base, dark);
		else
			dng_spot_fix(base);
	}
	if (channel != -1) {
		dng_comp_ch(base, channel);
	}
	term_dng(dark);
	
	if (path_clut != NULL) {
		struct dng_clut *clut;
		
		if ((clut = dng_clut_load(path_clut)) == NULL)
			perror(path_clut), exit(1);
		dng_clut_apply(base, clut);
	}
	
	return base;
}
