/*
 * TIFF/Exif metadata loader
 *  Copyright (C) 2015 Noriaki TANAKA (dtana@startide.jp)
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * $Id: tiffmd.c,v 1.1 2016/01/30 16:32:37 dtana Exp $
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <inttypes.h>

#ifdef RASTER
#include <tiff.h>
#include "raster.h"
#else
#include "tiffmd.h"
#endif

struct tiffmd {
	unsigned tag, type;
	int n, xo;

	long pe, pd;
	size_t size;
	void  *data;

	struct tiffmd *subd;
};

#define TERMIFDE 0xffff
#define IFD      0xfffe

static int find_head(FILE *, int *);
static int hostbe(void);
static int getn(FILE *, int, int);
static int make_node(struct tiffmd *, long, long, int, FILE *);
static void *emalloc(void *, size_t);

struct tiffmd *
load_tiffmd(FILE *fp)
{
	int be, xo, k;
	long ph;
	struct tiffmd *p;
	
	rewind(fp);
	
	p = NULL;
	if ((ph = find_head(fp, &be)) < 0) {
		errno = EINVAL;
		goto error_exit;
	}
	xo = hostbe() != be;
	
	for (k = 0;; k++) {
		int n;
		long pd;

		if ((pd = getn(fp, xo, 4)) == 0)
			break;
		
		if (fseek(fp, ph + pd, SEEK_SET) == EOF)
			goto error_exit;
		if ((n = getn(fp, xo, 2)) < 0)
			goto error_exit;
		if (fseek(fp, n * 12, SEEK_CUR) == EOF)
			goto error_exit;
		
		if ((p = emalloc(p, sizeof(struct tiffmd) * (k + 2))) == NULL)
			goto error_exit;
		p[k + 1].tag = TERMIFDE;
		if (make_node(p + k, ph, pd, xo, fp) < 0)
			goto error_exit;
	}
	return p;
	
error_exit:
	term_tiffmd(p);
	return NULL;
}

void
term_tiffmd(struct tiffmd *p)
{
	int n;
	
	for (n = 0; p && p[n].tag != TERMIFDE; n++) {
		free(p[n].data);
		term_tiffmd(p[n].subd);
	}
	free(p);
}

int
tiffmd_xo(const struct tiffmd *p)
{
	return p->xo;
}

/* IO depend on byteorder */
static void
copy(void *d, const void *s, int bytes, int xo)
{
	unsigned char *dp;
	const unsigned char *sp;
	int k;
	
	for (k = 0, dp = d, sp = s; k < bytes; k++)
		dp[k] = sp[xo ? (bytes - 1 - k): k];
}

static uint16_t
u16(void *p, int xo)
{
	uint16_t r;
	
	copy(&r, p, sizeof(r), xo);
	return r;
}

static uint32_t
u32(void *p, int xo)
{
	uint32_t r;

	copy(&r, p, sizeof(r), xo);
	return r;
}

static int16_t
s16(void *p, int xo)
{
	int16_t r;
	
	copy(&r, p, sizeof(r), xo);
	return r;
}

static int32_t
s32(void *p, int xo)
{
	int32_t r;

	copy(&r, p, sizeof(r), xo);
	return r;
}

static int
getn(FILE *fp, int xo, int bytes)
{
	char b[4];
	
	if (fread(b, bytes, 1, fp) != 1)
		return -1;
	return bytes == 4 ? u32(b, xo): u16(b, xo);
}

/* TIFF data types */
static const struct {
	int type;
	const char *name;
	size_t size;
} types[] = {
	{TIFF_NOTYPE,    "NOTYPE",	0},
	{TIFF_BYTE,      "BYTE",	1},
	{TIFF_ASCII,     "ASCII",	1},
	{TIFF_SHORT,     "SHORT",	2},
	{TIFF_LONG,      "LONG",	4},
	{TIFF_RATIONAL,  "RATIONAL",    8},
	{TIFF_SBYTE,     "SBYTE",	1},
	{TIFF_UNDEFINED, "UNDEFINED",   1},
	{TIFF_SSHORT,    "SSHORT",      2},
	{TIFF_SLONG,     "SLONG",	4},
	{TIFF_SRATIONAL, "SRATIONAL",   8},
	{TIFF_FLOAT,     "FLOAT",	4},
	{TIFF_DOUBLE,    "DOUBLE",      8},
};

#define NTYPES ((int)(sizeof(types)/sizeof(types[0])))

static size_t
type_size(int type)
{
	int n;
	
	for (n = 0; n < NTYPES; n++)
		if (types[n].type == type)
			return types[n].size;
	return 0;
}

static const char *
type_name(int type)
{
	int n;
	
	for (n = 0; n < NTYPES; n++)
		if (types[n].type == type)
			return types[n].name;
	return NULL;
}

static int find_exif_on_jfif(FILE *);

static int
find_head(FILE* fp, int *be)
{
	const char mm[] = {'M', 'M', 0x00, 0x2a};
	const char ii[] = {'I', 'I', 0x2a, 0x00};
	long ph;
	char b[4];
	
	if ((ph = find_exif_on_jfif(fp)) < 0)
		return -1;
	if (fseek(fp, ph, SEEK_SET) == EOF)
		return -1;
	if (fread(b, 4, 1, fp) != 1)
		return -1;
	if (memcmp(b, mm, 4) != 0 && memcmp(b, ii, 4) != 0)
		return -1;
	*be = b[0] == 'M';
	return ph;
}

/* Search jfif APP1 Exif Segment */
#define MARKER(c)   (((c)&0xFF00)==0xFF00)
#define RESTART(c)  (0xFFD0<=(c)&&(c)<=0xFFD7)
#define SOI  0xFFD8
#define EOI  0xFFD9
#define APP1 0xFFE1
#define SOS  0xFFDA
#define DATA 0xFF00

static int
skip_scan_segment(FILE *fp)
{
	int c;
	
	while ((c = fgetc(fp)) != EOF) {
		if (c != 0xFF)
			continue;
		
		if ((c = fgetc(fp)) == EOF)
			break;
		if ((c |= 0xFF00) == DATA || RESTART(c))
			continue;
		
		/* next segment found */
		if (fseek(fp, -2, SEEK_CUR) == EOF)
			break;
		return 0;
	}
	return -1;
}

#define up(p)	 ((unsigned char *)(p))
#define u16be(p) ((uint16_t)up(p)[1]|up(p)[0] << 8)

static int
get16be(FILE *fp)
{
	unsigned char b[2];

	if (fread(b, 2, 1, fp) != 1)
		return EOF;
	return u16be(b);
}

static int
find_exif_on_jfif(FILE *fp)
{
	int marker;
	
	rewind(fp);
	
	if (get16be(fp) != SOI)
		return 0;
	
	while ((marker = get16be(fp)) != EOF) {
		int size;
		long cp;
		
		/* printf("%04x: \n", marker); */
		if (!MARKER(marker))
			break;
		if (marker == EOI)
			return -1; /* no Exif entries */
		if (marker == SOS) {
			if (skip_scan_segment(fp) < 0)
				break;
			continue;
		}
		
		cp = ftell(fp);
		if ((size = get16be(fp)) < 2)
			break;
		
		if (marker == APP1) {
			char b[6];
			
			if (fread(b, 6, 1, fp) == 1 &&
			    memcmp(b, "Exif\0\0", 6) == 0)
				return ftell(fp);
		}
		if (fseek(fp, cp + size, SEEK_SET) == EOF)
			break;
	}
	return -1; /* incorrect JFIF */
}

#define EXIFIFDPOINTER             0x8769
#define GPSINFO                    0x8825
#define INTEROPERABILITYIFDPOINTER 0xa005
#define SUBIFDS                    0x014a
#define NEWSUBFILETYPE             0x00fe
#define FNUMBER                    0x829d
#define EXPOSURETIME               0x829a
#define FOCALLENGTH                0x920a
#define IMAGEDESCRIPTION           0x010e
#define DNGVERSION                 0xc612
#define DNGBACKWARDVERSION         0xc613

#define NODE(t) (					\
		(t)==EXIFIFDPOINTER||			\
		(t)==GPSINFO||				\
		(t)==INTEROPERABILITYIFDPOINTER ||	\
		(t)==SUBIFDS)

static struct tiffmd *read_list(FILE *, long, long, int);

static int
make_node(struct tiffmd *p, long ph, long pd, int xo, FILE *fp)
{
	uint32_t pd_;
	
	p->pe   = 0; /* pseudo entry */
	p->tag  = IFD;
	p->type = TIFF_LONG;
	p->n    = 1;
	p->pd   = pd;
	p->xo   = xo;
	p->size = 4;

	if ((p->data = malloc(p->size)) == NULL)
		return -1;
	
	pd_ = pd;
	copy(p->data, &pd_, sizeof(pd_), xo);
	
	if ((p->subd = read_list(fp, ph, pd, xo)) == NULL)
		return -1;
	return 0;
}

static int read_ifde(struct tiffmd *, long, int, FILE *);

static struct tiffmd *
read_list(FILE* fp, long ph, long pd, int xo)
{
	int k, n;
	struct tiffmd *p;
	long cp;
	
	cp = ftell(fp);
	if (fseek(fp, ph + pd, SEEK_SET) == EOF || (n = getn(fp, xo, 2)) < 0)
		return NULL;
	if ((p = malloc(sizeof(struct tiffmd) * (n + 1))) == NULL)
		return NULL;
	p[0].tag = TERMIFDE;
	
	for (k = 0; k < n; k++) {
		p[k].data = p[k].subd = NULL;
		if (read_ifde(p + k, ph, xo, fp) < 0)
			goto error_exit;
		p[k + 1].tag = TERMIFDE;
	}
	if (fseek(fp, cp, SEEK_SET) == EOF)
		goto error_exit;
	return p;

error_exit:
	term_tiffmd(p);
	return NULL;
}

static long tiff_long(const struct tiffmd *p, int k);

static int
read_ifde(struct tiffmd *p, long ph, int xo, FILE* fp)
{
	char b[12];
	int j;	
	long cp;

	p->pe = ftell(fp);
	if (fread(b, 12, 1, fp) != 1)
		return -1;
	p->xo = xo;
	
	p->tag  = u16(b,     xo);
	p->type = u16(b + 2, xo);
	p->n    = u32(b + 4, xo);
	p->subd = NULL;
	p->data = NULL;
	
	if ((p->size = type_size(p->type) * p->n) == 0)
		return 0;
	if ((p->data = malloc(p->size)) == NULL)
		return -1;

	if (4 < p->size)
		p->pd = u32(b + 8, xo) + ph;
	else
		p->pd = p->pe + 8;

	cp = ftell(fp);
	if (fseek(fp, p->pd, SEEK_SET) == EOF)
		return -1;
	if (fread(p->data, p->size, 1, fp) != 1)
		return -1;
	if (fseek(fp, cp, SEEK_SET) == EOF)
		return -1;
	
	if (!NODE(p->tag))
		return 0;
	
	if ((p->subd = malloc(sizeof(struct tiffmd) * p->n + 1)) == NULL)
		return -1;
	
	for (j = 0; j < p->n; j++) {
		if (make_node(p->subd + j, ph, tiff_long(p, j), xo, fp) < 0)
			return -1;
		p->subd[j + 1].tag = TERMIFDE;
	}
	return 0;
}

static long
tiff_long(const struct tiffmd *p, int k)
{
	int r;
	
	if (tiffmd_getn(&r, 1, p, k) != 1)
		return 0;
	return (long)r;
}

static void *
emalloc(void *p, size_t size)
{
	return p == NULL ? malloc(size): realloc(p, size);
}

static int
hostbe(void)
{
	const int a = 1;
	return *(char *)&a != 1;
}

const struct tiffmd *
find_tiffmd(const struct tiffmd *p, unsigned tag)
{
	for (; p && p->tag != TERMIFDE; p++) {
		const struct tiffmd *sp;
		
		if (p->tag == tag)
			return p;
		if ((sp = find_tiffmd(p->subd, tag)) != NULL)
			return sp;
	}
	return NULL;
}

const struct tiffmd *
find_tiffmd_node(const struct tiffmd *p, unsigned tag, int n)
{
	if (tag == IFD) {
		int k;

		for (k = 0; p && p->tag != TERMIFDE; k++, p++)
			if (k == n)
				return p->subd;
		return NULL;
	}
	if ((p = find_tiffmd(p, tag)) == NULL)
		return NULL;
	return find_tiffmd_node(p->subd, IFD, n);
}

int
tiffmd_getn(int d[], int n, const struct tiffmd *p, int k)
{
	char *tp;
	
	if (p->n <= k)
		return -1;
	tp = (char *)p->data + type_size(p->type) * k;

	switch (p->type) {
	case TIFF_BYTE:
		d[0] = *(uint8_t *)tp;
		return 1;
	case TIFF_SBYTE:
		d[0] = *(int8_t *)tp;
		return 1;
	case TIFF_SHORT:
		d[0] = u16(tp, p->xo);
		return 1;
	case TIFF_SSHORT:
		d[0] = s16(tp, p->xo);
		return 1;
	case TIFF_LONG:
		d[0] = u32(tp, p->xo);
		return 1;
	case TIFF_SLONG:
		d[0] = s32(tp, p->xo);
		return 1;
	case TIFF_RATIONAL:
		if (n != 2)
			return -1;
		d[0] = u32(tp,     p->xo);
		d[1] = u32(tp + 4, p->xo);
		return 2;
	case TIFF_SRATIONAL:
		if (n != 2)
			return -1;
		d[0] = s32(tp,     p->xo);
		d[1] = s32(tp + 4, p->xo);
		return 2;
	default:
		return -1;
	}
	return -1;
}

int
tiffmd_setn(const int d[], int n, struct tiffmd *p, int k)
{
	char *tp;
	uint16_t uw;
	uint32_t ud;
	int16_t  sw;
	int32_t  sd;
	
	if (p->n <= k)
		return -1;
	tp = (char *)p->data + type_size(p->type) * k;
	
	switch (p->type) {
	case TIFF_BYTE:
		*(uint8_t *)tp = (uint8_t)d[0];
		return 1;
	case TIFF_SBYTE:
		*(int8_t *)tp  =  (int8_t)d[0];
		return 1;
	case TIFF_SHORT:
		uw = (uint16_t)d[0], copy(tp, &uw, 2, p->xo);
		return 1;
	case TIFF_SSHORT:
		sw =  (int16_t)d[0], copy(tp, &sw, 2, p->xo);
		return 1;
	case TIFF_LONG:
		ud = (uint32_t)d[0], copy(tp, &ud, 4, p->xo);
		return 1;
	case TIFF_SLONG:
		sd =  (int32_t)d[0], copy(tp, &sd, 4, p->xo);
		return 1;
	case TIFF_RATIONAL:
		if (n != 2)
			return -1;
		ud = (uint32_t)d[0], copy(tp,     &ud, 4, p->xo);
		ud = (uint32_t)d[1], copy(tp + 4, &ud, 4, p->xo);
		return 2;
	case TIFF_SRATIONAL:
		if (n != 2)
			return -1;
		sd =  (int32_t)d[0], copy(tp,     &sd, 4, p->xo);
		sd =  (int32_t)d[1], copy(tp + 4, &sd, 4, p->xo);
		return 2;
	default:
		return -1;
	}
	return -1;
}

int
tiffmd_getd(char *b, size_t size, const struct tiffmd *p)
{
	if (b == NULL)
		return (int)p->size;

	if (p->size < size)
		size = p->size;
	memcpy(b, p->data, size);
	return (int)size;
}

int
tiffmd_pute(const struct tiffmd *p, char *data, size_t size)
{
	char b[12];
	uint16_t w;
	uint32_t d;
	
	if (size < (size_t)p->pe + 12 ||
	    size < (size_t)p->pd + p->size) {
		fprintf(stderr, "pointers overflow\n");
		return -1;
	}
	w = p->tag,  copy(b    , &w, 2, p->xo);
	w = p->type, copy(b + 2, &w, 2, p->xo);
	d = p->n,    copy(b + 4, &d, 4, p->xo);
	if (4 < p->size)
		d = p->pd, copy(b + 8, &d, 4, p->xo);
	memcpy(data + p->pe, b, 12);
	memcpy(data + p->pd, p->data, p->size);

	return 0;
}

void
tiffmd_set_disable(struct tiffmd *p)
{
	char *data;
	
	p->tag  = IMAGEDESCRIPTION;
	p->type = TIFF_ASCII;

	data = p->data;
	data[0] = '*';
	data[1] = '\0';
}


static struct {
	unsigned tag;
	const char *desc;
} tags[] = {
#include "tags.h"	
};

static const char *
tag_to_str(unsigned tag)
{
	int k, n;

	n = sizeof(tags)/sizeof(tags[0]);
	for (k = 0; k < n; k++)
		if (tags[k].tag == tag)
			return tags[k].desc;
	return NULL;
}

static int
matchstr(const char *a, const char *b)
{
	for (; *a && *b; a++, b++)
		if (toupper(*a) != toupper(*b))
			return 0;
	return *a == 0 && *b == 0;
}

int
tiffmd_tag(const char *name)
{
	int k, n;
	
	n = sizeof(tags)/sizeof(tags[0]);
	for (k = 0; k < n; k++)
		if (matchstr(tags[k].desc, name))
			return (int)tags[k].tag;
	return -1;
}

void
dump_tiffmd(char *b, int n, const struct tiffmd *p)
{
	int r[2], rn;
	
	if (n == 0)
		return;

	b[0] = 0;
	if (p == NULL)
		return;
	if (p->type == TIFF_ASCII) {
		tiffmd_getd(b, n, p), b[n - 1] = 0;
		return;
	}
	if ((rn = tiffmd_getn(r, 2, p, 0)) == 1) {
		sprintf(b, "%d", r[0]);
		if (p->tag == NEWSUBFILETYPE) {
			sprintf(b + strlen(b), " %s", 
			    (r[0]&0x1) ? "REDUCEDIMAGE": "");
			return;
		}
		if (p->tag == DNGVERSION ||
		    p->tag == DNGBACKWARDVERSION) {
			int k;
			for (k = 1; k < p->n; k++) {
				tiffmd_getn(r, 1, p, k);
				sprintf(b + strlen(b), " %d", r[0]);
			}
			return;
		}
		return;
	}
	if (rn == 2) {
		int nd, dd;
		
		nd = r[0];
		dd = r[1];
		
		if (p->tag == FNUMBER && dd != 0) {
			const char *fs;
			double fn;
			
			fn = (double)nd / dd;
			fs = (fn >= 10.0 || nd % dd == 0) ?
			    "%.0f": "%.1f";
			sprintf(b, fs, fn);
			
			return;
		}
		if (p->tag == EXPOSURETIME && dd != 0) {
			const char *fs;
			double et;
			
			et = (double)nd / dd;
			if (1.0 <= et) {
				fs = (et >= 10.0 || nd % dd == 0) ?
				    "%.0f": "%.1f";
				sprintf(b, fs, et);
				return;
			}
		}
		if (p->tag == FOCALLENGTH && dd != 0) {
			const char *fs;
			double fl;
			
			fl = (double)nd / dd;
			fs = (fl >= 20.0 || nd % dd == 0) ?
			    "%.0f": "%.1f";
			sprintf(b, fs, fl);
			
			return;
		}

		if (nd != 0 && dd % nd == 0)
			dd = dd / nd, nd = 1;
		sprintf(b, "%d/%d", nd, dd);
		return;
	}
	{
		const unsigned char *r = p->data;
		sprintf(b, "%02x %02x %02x %02x", r[0], r[1], r[2], r[3]);
	}
}

static void
dump(const struct tiffmd *p, int h)
{
	char t[BUFSIZ];
	const char *s;

	if ((s = tag_to_str(p->tag)) == NULL)
		s = t, sprintf(t, "%04x", p->tag);
	printf("%*s", h, "");
	printf("%-20.20s: %9s: %4d: ", s, type_name(p->type), p->n);

	if (p->size <= 4)
		printf("--------: ");
	else
		printf("%8d: ", (int)p->pd);
	
	dump_tiffmd(t, BUFSIZ, p);
	printf("%s\n", t);
}

int
list_tiffmd(const struct tiffmd *p, int h)
{
	for (; p && p->tag != TERMIFDE; p++) {
		dump(p, h);
		list_tiffmd(p->subd, h + (p->tag == IFD ? 1:2));
	}
	return 0;
}
